/*******************************************************************
  uLan Communication - basic library

  ul_l_drv.c	- driver interface

  (C) Copyright 1996,1999 by Pavel Pisa 

  The uLan driver is distributed under the Gnu General Public Licence. 
  See file COPYING for details.
 *******************************************************************/

#include <sys/types.h>
#ifndef _MSC_VER
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#endif /*_MSC_VER*/
#include <ul_lib/ulan.h>

#if !defined(_WIN32) && !defined(UL_DRV_IN_LIB)

inline ul_fd_t ul_open(const char *dev_name, const char *options)
{
  if(dev_name==NULL) dev_name=UL_DEV_NAME;
  return open(dev_name,O_RDWR, S_IWRITE | S_IREAD);
}

inline int ul_close(ul_fd_t ul_fd)
{
  return close(ul_fd);
}

inline ssize_t ul_read(ul_fd_t ul_fd, void *buffer, size_t size)
{
  return read(ul_fd,buffer,size);
}

inline ssize_t ul_write(ul_fd_t ul_fd, const void *buffer, size_t size)
{
  return write(ul_fd,buffer,size);
}

inline int ul_newmsg(ul_fd_t ul_fd,const ul_msginfo *msginfo)
{
  return ioctl(ul_fd,UL_NEWMSG,msginfo);
}

inline int ul_tailmsg(ul_fd_t ul_fd,const ul_msginfo *msginfo)
{
  return ioctl(ul_fd,UL_TAILMSG,msginfo);
}

inline int ul_freemsg(ul_fd_t ul_fd)
{
  return ioctl(ul_fd,UL_FREEMSG);
}

inline int ul_acceptmsg(ul_fd_t ul_fd,ul_msginfo *msginfo)
{
  return ioctl(ul_fd,UL_ACCEPTMSG,msginfo);
}

inline int ul_actailmsg(ul_fd_t ul_fd,ul_msginfo *msginfo)
{
  return ioctl(ul_fd,UL_ACTAILMSG,msginfo);
}

inline int ul_addfilt(ul_fd_t ul_fd,const ul_msginfo *msginfo)
{
  return ioctl(ul_fd,UL_ADDFILT,msginfo);
}

inline int ul_abortmsg(ul_fd_t ul_fd)
{
  return ioctl(ul_fd,UL_ABORTMSG);
}

inline int ul_rewmsg(ul_fd_t ul_fd)
{
  return ioctl(ul_fd,UL_REWMSG);
}

inline int ul_inepoll(ul_fd_t ul_fd)
{
  return ioctl(ul_fd,UL_INEPOLL);
}

inline int ul_drv_debflg(ul_fd_t ul_fd,int debug_msk)
{
  return ioctl(ul_fd,UL_DEBFLG,debug_msk);
}

int ul_fd_wait(ul_fd_t ul_fd, int wait_sec)
{
  int ret;
  struct timeval timeout;
  fd_set set;

  FD_ZERO (&set);
  FD_SET (ul_fd, &set);
  timeout.tv_sec = wait_sec;
  timeout.tv_usec = 0;
  while ((ret=select(FD_SETSIZE,&set, NULL, NULL,&timeout))==-1
          &&errno==-EINTR);
  return ret;
}

/*******************************************************************/
#elif defined(_WIN32)

ul_fd_t ul_open(const char *dev_name, const char *options)
{
  if(dev_name==NULL) dev_name=UL_DEV_NAME;
  return CreateFile(dev_name,GENERIC_READ | GENERIC_WRITE,
                    0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
}

int ul_close(ul_fd_t ul_fd)
{
  if(!CloseHandle(ul_fd)) return -1;
  return 0;
}

ssize_t ul_read(ul_fd_t ul_fd, void *buffer, size_t size)
{
  DWORD ret;
  if(!ReadFile(ul_fd,buffer,size,&ret,NULL)) return -1;
  return ret;
}

ssize_t ul_write(ul_fd_t ul_fd, const void *buffer, size_t size)
{
  DWORD ret;
  if(!WriteFile(ul_fd,buffer,size,&ret,NULL)) return -1;
  return ret;
}

int ul_newmsg(ul_fd_t ul_fd,const ul_msginfo *msginfo)
{
  DWORD bytes_ret;
  if(!DeviceIoControl(ul_fd,UL_NEWMSG,
                      (void *)msginfo,sizeof(ul_msginfo),NULL,0,
                      &bytes_ret,NULL)) return -1;
  return 0;
}

int ul_tailmsg(ul_fd_t ul_fd,const ul_msginfo *msginfo)
{
  DWORD bytes_ret;
  if(!DeviceIoControl(ul_fd,UL_TAILMSG,
                      (void *)msginfo,sizeof(ul_msginfo),NULL,0,
                      &bytes_ret,NULL)) return -1;
  return 0;
}

int ul_freemsg(ul_fd_t ul_fd)
{
  DWORD bytes_ret;
  int ret;
  if(!DeviceIoControl(ul_fd,UL_FREEMSG,NULL,0,&ret,sizeof(ret),
                      &bytes_ret,NULL)) return -1;
  if(bytes_ret!=sizeof(ret)) return -1;
  return ret;
}

int ul_acceptmsg(ul_fd_t ul_fd,ul_msginfo *msginfo)
{
  DWORD bytes_ret;
  if(!DeviceIoControl(ul_fd,UL_ACCEPTMSG,
                      NULL,0,msginfo,sizeof(ul_msginfo),
                      &bytes_ret,NULL)) return -1;
  if(bytes_ret!=sizeof(ul_msginfo)) return -1;
  return 0;
}

int ul_actailmsg(ul_fd_t ul_fd,ul_msginfo *msginfo)
{
  DWORD bytes_ret;
  if(!DeviceIoControl(ul_fd,UL_ACTAILMSG,
                      NULL,0,msginfo,sizeof(ul_msginfo),
                      &bytes_ret,NULL)) return -1;
  if(bytes_ret!=sizeof(ul_msginfo)) return -1;
  return 0;
}

int ul_addfilt(ul_fd_t ul_fd,const ul_msginfo *msginfo)
{
  DWORD bytes_ret;
  if(!DeviceIoControl(ul_fd,UL_ADDFILT,
                      (void *)msginfo,sizeof(ul_msginfo),NULL,0,
                      &bytes_ret,NULL)) return -1;
  return 0;
}

int ul_abortmsg(ul_fd_t ul_fd)
{
  DWORD bytes_ret;
  if(!DeviceIoControl(ul_fd,UL_ABORTMSG,NULL,0,NULL,0,
                      &bytes_ret,NULL)) return -1;
  return 0;
}

int ul_rewmsg(ul_fd_t ul_fd)
{
  DWORD bytes_ret;
  if(!DeviceIoControl(ul_fd,UL_REWMSG,NULL,0,NULL,0,
                      &bytes_ret,NULL)) return -1;
  return 0;
}

int ul_inepoll(ul_fd_t ul_fd)
{
  DWORD bytes_ret;
  int ret;
  if(!DeviceIoControl(ul_fd,UL_INEPOLL,NULL,0,&ret,sizeof(ret),
                      &bytes_ret,NULL)) return -1;
  if(bytes_ret!=sizeof(ret)) return -1;
  return ret;
}

int ul_drv_debflg(ul_fd_t ul_fd,int debug_msk)
{
  DWORD bytes_ret;
  if(!DeviceIoControl(ul_fd,UL_DEBFLG,
		  &debug_msk,sizeof(debug_msk),NULL,0,
		  &bytes_ret,NULL)) return -1;
  return 0;
}

#ifndef WITHOUT_SYS_SELECT

int ul_fd_wait(ul_fd_t ul_fd, int wait_sec)
{
  int ret;
  struct timeval timeout;
  fd_set set;

  FD_ZERO (&set);
  FD_SET (ul_fd, &set);
  timeout.tv_sec = wait_sec;
  timeout.tv_usec = 0;
  while ((ret=select(FD_SETSIZE,&set, NULL, NULL,&timeout))==-1
          &&errno==-EINTR);
  return ret;
}

#else /* WITHOUT_SYS_SELECT */

int ul_fd_wait(ul_fd_t ul_fd, int wait_sec)
{
  int ret;
  long loops=0;
  do
  { ret=ul_inepoll(ul_fd);
    if(ret) return ret;
    Sleep(100);
  }while(loops++<(wait_sec*10));
 
  return 0;
}

#endif /* WITHOUT_SYS_SELECT */

#elif defined(__DJGPP__)

int ul_fd_wait(ul_fd_t ul_fd, int wait_sec)
{
  int ret;
  long loops=0;
  long int starttime;
  struct timeval actual;
  gettimeofday(&actual,NULL);
  starttime=actual.tv_sec+1;
  do
  { ret=ul_inepoll(ul_fd);
    if(ret) return ret;
    /* uld_djgpp_call_irq(); */
    gettimeofday(&actual,NULL);
    loops++;
  }while((wait_sec+starttime-(long)actual.tv_sec)>0);
 
  return 0;
}

#elif defined(UL_DRV_IN_LIB)

int ul_fd_wait(ul_fd_t ul_fd, int wait_sec)
{
  int ret;
  long loops=0;
  do
  { ret=ul_inepoll(ul_fd);
    if(ret) return ret;
    usleep(100*1000);
    /* uld_djgpp_call_irq(); */
  }while(loops++<(wait_sec*10));
 
  return 0;
}

#else
  #error Interface to driver not defined for target
#endif
